part of modulation;

abstract class Modulator {
  
  static const double MAX_PHASE_CYCLE = 1.0;
  static const double MIN_PHASE_CYCLE_NEG = -1.0;
  
  double phase = 0.0, 
          phaseStep = 0.0,
          frequency = 0.0;
  
  // ================
  Modulator(frequency) { setFrequency(frequency); }
  
  // ================
  void setFrequency(double newFrequency) {
    frequency = newFrequency;
    phaseStep = MAX_PHASE_CYCLE * frequency;
  }
  
  // ================
  double nextPhase() {
    phase += phaseStep;
    if (phase > MAX_PHASE_CYCLE)
      // reset phase and add any excess remainder
      phase = MIN_PHASE_CYCLE_NEG + (phase - MAX_PHASE_CYCLE);
    
    return phase;
  }

  // ================
  @override
  double get current;
  
  // ================
  void clear() {
    phase = 0.0;
    phaseStep = 0.0;
    frequency = 0.0;
  }
  
}

class SineModulator extends Modulator {
  
  // ================
  SineModulator(frequency) : super(frequency);
  
  // ================
  @override
  double get current => Math.sin(phase * Math.PI);
  
}

class SawModulator extends Modulator {
  
  // ================
  SawModulator(frequency) : super(frequency);
  
  // ================
  @override
  double get current => phase/Modulator.MAX_PHASE_CYCLE;
  
}

class SquareModulator extends Modulator {
  
  double _current;
  
  // ================
  SquareModulator(frequency) : super(frequency) {
    _current = Modulator.MAX_PHASE_CYCLE;
  }
  
  // ================
  @override
  double nextPhase() {
    phase += phaseStep;
    if (phase > Modulator.MAX_PHASE_CYCLE) {
      phase = 0.0;
      _current = -_current;
    }
    
    return phase;
  }
  
  // ================
  @override
  double get current => _current;
  
}

class TriModulator extends Modulator {
  
  static const double TRI_SQUARE_MAX_CYCLE = Modulator.MAX_PHASE_CYCLE / 2;
  static const double TRI_SQUARE_MAX_CYCLE_NEG = -TRI_SQUARE_MAX_CYCLE;
  
  // ================
  TriModulator(frequency) : super(frequency);
  
  // ================
  @override
  double nextPhase() {
    phase += phaseStep;
    
    if (phase > TRI_SQUARE_MAX_CYCLE) {
      // reset phase and add any excess remainder
      phase = TRI_SQUARE_MAX_CYCLE - (phase - TRI_SQUARE_MAX_CYCLE);
      phaseStep = -phaseStep;
    } else if (phase < TRI_SQUARE_MAX_CYCLE_NEG) {
      // reset phase and add any excess remainder
      phase = TRI_SQUARE_MAX_CYCLE_NEG - (phase - TRI_SQUARE_MAX_CYCLE_NEG);
      phaseStep = -phaseStep;
    }
    
    return phase;
  }

  // ================
  @override
  double get current => (phase/TRI_SQUARE_MAX_CYCLE);
  
}
